<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2020.1/2030377
 */

namespace Reviews\Model;

use Application\Config\IDao;
use Application\Config\ConfigManager;
use Application\Config\Services;
use Application\Config\ConfigException;
use Application\Connection\ConnectionFactory;
use Application\Filter\FilterException;
use Application\I18n\TranslatorFactory;
use Application\Log\SwarmLogger;
use Application\Model\AbstractDAO;
use Application\Model\IModelDAO;
use Application\Option;
use Application\Permissions\Exception\ForbiddenException;
use Application\Permissions\Permissions;
use Application\Permissions\PrivateProjects;
use Application\Permissions\Reviews;
use Events\Listener\ListenerFactory;
use Files\View\Helper\DecodeSpec;
use InvalidArgumentException;
use Groups\Model\Config as GroupConfig;
use P4\Connection\ConnectionInterface;
use P4\Exception as P4Exception;
use P4\Spec\Change;
use P4\Spec\Exception\NotFoundException as SpecNotFoundException;
use Projects\Model\Project as ProjectModel;
use Record\Exception\Exception;
use Record\Exception\NotFoundException as RecordNotFoundException;
use Reviews\Filter\GitInfo;
use Reviews\Filter\Keywords;
use Reviews\Filter\Participants as ParticipantsFilter;
use Reviews\Listener\IReviewTask;
use Reviews\Model\Review as ReviewModel;
use Reviews\Filter\IParticipants;
use Reviews\UpdateService;
use Users\Validator\Users;
use Reviews\ITransition;
use Reviews\Validator\Transitions;
use Workflow\Model\IWorkflow;
use Laminas\Validator\NotEmpty;
use Laminas\Validator\ValidatorChain;
use Queue\Manager as QueueManager;
use Laminas\Http\Request as HttpRequest;

class ReviewDAO extends AbstractDAO implements ITransition, IReviewTask
{
    // The Perforce class that handles review
    const MODEL = ReviewModel::class;
    // Version is used in the DAO but is not actually a field on the model
    const VERSION = 'version';

    /**
     * @inheritDoc
     * @throws ForbiddenException
     * @throws RecordNotFoundException
     */
    public function fetch($id, ConnectionInterface $connection = null)
    {
        $review = ReviewModel::fetch($id, $this->getConnection($connection));
        // checkAccess will result in a ForbiddenException if no access any projects on the review is detected.
        // This must be done before filterProjects so that it has projects to check
        $this->checkAccess($review);
        return $this->filterProjects($review);
    }

    /**
     * Fetches a review base on keywords from value using the standard Swarm definition (for example #review-123) etc.
     * @param string    $value      value containing the keywords to find the review
     * @param ConnectionInterface|null $connection
     * @throws ForbiddenException
     * @throws RecordNotFoundException
     * @return mixed the review if keywords are present and the review is found, null if there are no keywords to find a
     * review, and a RecordNotFoundException if the keywords specify a review that does not exist
     */
    public function fetchFromKeyword($value, ConnectionInterface $connection = null)
    {
        $keywords = $this->services->get(Keywords::SERVICE);
        $matches  = $keywords->getMatches($value);
        if ($matches && isset($matches['id']) && !empty($matches['id'])) {
            return $this->fetch($matches['id'], $connection);
        }
        return null;
    }

    /**
     * Filter private projects from the review if there are projects set on the review
     * @param mixed     $review     the review
     * @return mixed the review with private projects removed
     */
    public function filterProjects($review)
    {
        $projects = $review->getProjects();
        $retVal   = $review;
        if ($projects) {
            $retVal = $review->setProjects(
                $this->services->get(PrivateProjects::PROJECTS_FILTER)->filterList($projects)
            );
        }
        return $retVal;
    }

    /**
     * Wrapping the obliterate function of review model
     *
     * @param   mixed   $review            Review Object.
     * @param   bool    $removeChangelist  True, will remove the pending changelist.
     * @return  array
     * @throws  Exception
     */
    public function obliterate(ReviewModel $review, $removeChangelist = true)
    {
        return $review->obliterate($removeChangelist);
    }

    /**
     * Change the author of a review to a new value.
     * The change will only be effected if:
     *  The Swarm config allows author changes to be made
     *  The review exists
     *  The current user is authorised to update this review
     *  The new author is a valid user
     * @param $reviewId
     * @param $userId
     * @param $data
     * @return mixed
     * @throws ConfigException
     * @throws ForbiddenException
     * @throws RecordNotFoundException
     */
    public function changeAuthor($reviewId, $userId, $data)
    {
        $services = $this->services;
        $config   = $services->get(ConfigManager::CONFIG);
        if (!ConfigManager::getValue($config, ConfigManager::REVIEWS_ALLOW_AUTHOR_CHANGE)) {
            // Allow author change is not enabled
            throw new \InvalidArgumentException(
                $services->get(TranslatorFactory::SERVICE)->t(
                    "Review author cannot be changed, please set 'allow_author_change' to be true."
                )
            );
        }
        $p4Admin = $services->get(ConnectionFactory::P4_ADMIN);
        $review  = $this->fetch($reviewId, $p4Admin);
        $this->checkAccess($review);
        $userValidator = (new ValidatorChain())->attach(new NotEmpty())->attach(new Users(['connection' => $p4Admin]));
        $newAuthor     = $data[Review::FIELD_AUTHOR];
        if ($userValidator->isValid($newAuthor)) {
            $oldValues = $review->get();
            if ($this->hasChanged(Review::FIELD_AUTHOR, [Review::FIELD_AUTHOR => $newAuthor], $oldValues)) {
                $review->set(Review::FIELD_AUTHOR, $newAuthor);
                $review = $this->save($review);
                $this->queueTask($review, [self::IS_AUTHOR_CHANGE => true], $oldValues);
            }
        } else {
            $messages = $userValidator->getMessages();
            throw new \InvalidArgumentException(implode(', ', array_values($messages)));
        }
        return $data;
    }

    /**
     * @param mixed $review
     * @throws ForbiddenException
     */
    public function checkAccess($review)
    {
        if (!$this->services->get(Reviews::REVIEWS_FILTER)->canAccessChangesAndProjects($review)) {
            $translator = $this->services->get(TranslatorFactory::SERVICE);
            throw new ForbiddenException(
                $translator->t('You do not have permission to access this review.')
            );
        }
    }

    /**
     * Vote on a review
     * @param mixed         $reviewId   the review id
     * @param mixed         $userId     the user id
     * @param array         $data       data, for example ['vote' => 'up']
     * @return array
     * @throws ConfigException
     * @throws FilterException
     * @throws ForbiddenException
     * @throws P4Exception
     * @throws RecordNotFoundException
     * @throws SpecNotFoundException
     */
    public function vote($reviewId, $userId, $data)
    {
        $review = $this->fetch($reviewId, $this->services->get(ConnectionFactory::P4_ADMIN));
        $this->checkAccess($review);
        $filter = $this->services->get(Services::VOTE_INPUT_FILTER);
        $filter->setData($data);
        if ($filter->isValid()) {
            $version     = isset($data[self::VERSION]) ? $data[self::VERSION] : null;
            $headVersion = $review->getHeadVersion();
            if ($version && (int)$version !== (int)$headVersion) {
                $translator = $this->services->get(TranslatorFactory::SERVICE);
                throw new InvalidArgumentException(
                    $translator->t(
                        'Version [%s] is not valid, must equal the head revision [%s].',
                        [
                            $version,
                            $headVersion
                        ]
                    )
                );
            }
            $oldValues = $review->get();
            $data      = $filter->getValues();
            $review    = $review->addParticipant($userId)
                ->addVote($userId, $data[ReviewModel::FIELD_VOTE], $headVersion);
            if ($this->canApprove($review)) {
                $review = $review->setState(ReviewModel::STATE_APPROVED);
            }
            $review = $this->save($review);
            $this->queueTask($review, [self::IS_VOTE => true], $oldValues);
        } else {
            throw new FilterException($filter);
        }
        return [Review::FIELD_VOTE=>$review->getVotes()[$userId]??[]];
    }

    /**
     * Tests if a review can be approved by looking at moderators and workflow rules
     * @param ReviewModel $review the review
     * @return bool true if the review is ready to be approved
     * @throws P4Exception
     * @throws SpecNotFoundException
     * @throws RecordNotFoundException
     */
    public function canApprove(ReviewModel $review)
    {
        $canApprove = false;
        $p4admin    = $this->services->get(ConnectionFactory::P4_ADMIN);
        $moderators = false;
        $affected   = $review->getProjects();
        $projectDAO = $this->services->get(IModelDAO::PROJECT_DAO);
        $projects   = $projectDAO->fetchAll(
            [ProjectModel::FETCH_BY_IDS => array_keys($affected)],
            $p4admin
        );
        foreach ($projects as $project) {
            if (count($project->getModerators($affected[$project->getId()])) > 0) {
                // We have at least one moderator
                $moderators = true;
                break;
            }
        }
        if (!$moderators) {
            $workflowsEnabled =
                $this->services->get(Services::CONFIG_CHECK)->check(IWorkflow::WORKFLOW_CHECKER_RETURN);
            if ($workflowsEnabled) {
                $workflowManager = $this->services->get(Services::WORKFLOW_MANAGER);
                // There are no moderators, look for a workflow with auto-approve = 'votes'
                $autoApproveRule = $workflowManager->getBranchRule(
                    IWorkflow::AUTO_APPROVE,
                    $affected,
                    $projects
                );
                $branchRules     = [];
                if ($autoApproveRule === IWorkflow::VOTES &&
                    $review->canApprove([], $affected, $workflowsEnabled, $branchRules)) {
                    $canApprove = true;
                }
            }
        }
        return $canApprove;
    }

    /**
     * Get allowed transitions for the review
     * The change will only be effected if:
     *  The review exists
     *  The user (if provided) is allowed the access the review. If the user is anonymous (permitted) this check will
     *  still prevent access if private projects on the review are encountered
     * @param mixed     $reviewId   the review id
     * @param mixed     $userId     the user id
     * @return array|false transitions keyed on transition id with a value of transition description, or false if
     * there are impacted projects but the requester is not a member/author/super
     * @throws ForbiddenException
     * @throws RecordNotFoundException
     */
    public function getTransitions($reviewId, $userId)
    {
        $review = $this->fetch($reviewId, $this->services->get(ConnectionFactory::P4_ADMIN));
        $this->checkAccess($review);
        $options     = [
            Option::USER_ID           => $userId,
            Transitions::OPT_UP_VOTES => $userId ? [$userId] : [],
            Transitions::REVIEW       => $review
        ];
        $transitions = $this->services->build(Services::TRANSITIONS, $options);
        return $transitions->getAllowedTransitions();
    }

    /**
     * Transition the review with the requested data
     * @param mixed         $reviewId   the review id
     * @param string        $userId     the user id
     * @param array         $data       the data
     * @see Transitions::transition()
     * @throws ForbiddenException
     * @throws RecordNotFoundException
     * @return array data with review state and state label
     */
    public function transition($reviewId, $userId, $data)
    {
        $review = $this->fetch($reviewId, $this->services->get(ConnectionFactory::P4_ADMIN));
        $this->checkAccess($review);
        $options   = [
            Option::USER_ID => $userId,
            Transitions::OPT_UP_VOTES => [$userId],
            Transitions::REVIEW => $review
        ];
        $validator = $this->services->build(Services::TRANSITIONS, $options);
        if (isset($data[self::TRANSITION])) {
            $oldValues = $review->get();
            $review    = $validator->transition($data);
            $this->queueTask($review, [self::IS_STATE_CHANGE => true], $oldValues);
        } else {
            $translator = $this->services->get(TranslatorFactory::SERVICE);
            throw new InvalidArgumentException($translator->t('No transition state provided'));
        }
        return [
            ReviewModel::FIELD_STATE => $review->getState(),
            ReviewModel::FIELD_STATE_LABEL => $review->getStateLabel()
        ];
    }

    /**
     * Function to work out if a fields have changed
     * @param string $field the review field
     * @param array $values current values
     * @param array $oldValues old values
     * @return bool true if changed
     */
    public function hasChanged(string $field, array $values, array $oldValues)
    {
        if (!array_key_exists($field, $values)) {
            return false;
        }
        $oldValues += [
            ReviewModel::FIELD_STATE         => null,
            ReviewModel::FIELD_AUTHOR        => null,
            ReviewModel::FIELD_TEST_STATUS   => null,
            ReviewModel::FIELD_DEPLOY_STATUS => null,
            ReviewModel::FIELD_DESCRIPTION   => null
        ];

        $newValue = $values[$field];
        $oldValue = isset($oldValues[$field]) ? $oldValues[$field] : null;
        // 'approved' to 'approved:commit' is not considered a state change
        if ($field === ReviewModel::FIELD_STATE && $newValue === ReviewModel::STATE_APPROVED_COMMIT) {
            $newValue = ReviewModel::STATE_APPROVED;
        }
        return $newValue != $oldValue;
    }

    /**
     * Revert a review to the values from the original state
     * @param ReviewModel       $review         the review
     * @param string            $originalState  original state
     * @return mixed
     */
    // Created from the revert handling in Reviews->IndexController->revertState and relocated
    public function revertState(ReviewModel $review, string $originalState)
    {
        try {
            $review->setState($originalState);
            $this->save($review);
        } catch (\Exception $e) {
            $logger = $this->services->get(SwarmLogger::SERVICE);
            $logger->err(sprintf('Unable to revert review %s to state %s', $review->getId(), $originalState));
        }
        return $review;
    }

    /**
     * Handle updating description on the original changelist or git info
     * @param ReviewModel $review
     * @param $data
     * @param $user
     * @return Review
     * @throws P4Exception
     * @throws SpecNotFoundException
     */
    // Created from the description handling in Reviews->IndexController->editReview
    public function updateDescription(ReviewModel $review, $data, $user)
    {
        $p4 = $review->getConnection();
        $p4->getService('clients')->grab();
        $change = Change::fetchById($review->getId(), $p4);

        // if this is a git review update the description but keep the
        // existing git info (keys/values), otherwise just use the review
        // description as-is for the update
        $description = $review->getDescription();
        if ($review->getType() === 'git') {
            $gitInfo     = new GitInfo($change->getDescription());
            $description = $gitInfo->setDescription($description)->getValue();
        }
        $changeDao = $this->services->get(IDao::CHANGE_DAO);
        $change    = $change->setDescription($description);
        $changeDao->save($change, true);
        $changeDao->updateOriginalChangelist($data, $review, $description, $user, $p4);
        $p4->getService('clients')->release();
        return $review;
    }

    /**
     * The change will only be effected if:
     *  The review exists
     *  The user is allowed the access the review.
     * @param mixed     $reviewId   the review id
     * @param array     $data       the data containing the description in $data[ReviewModel::FIELD_DESCRIPTION]
     * @throws ForbiddenException
     * @throws InvalidArgumentException
     * @throws P4Exception
     * @throws SpecNotFoundException
     * @throws RecordNotFoundException
     * @return array
     */
    public function setDescription($reviewId, $data)
    {
        $review = $this->fetch($reviewId, $this->services->get(ConnectionFactory::P4_ADMIN));
        $this->checkAccess($review);
        if (isset($data[ReviewModel::FIELD_DESCRIPTION])) {
            if ($this->hasChanged(
                ReviewModel::FIELD_DESCRIPTION,
                [ReviewModel::FIELD_DESCRIPTION => $data[ReviewModel::FIELD_DESCRIPTION]],
                [$review->getDescription()]
            )) {
                $oldValues = $review->get();
                $review    = $review->setDescription($data[ReviewModel::FIELD_DESCRIPTION]);
                $review    = $this->save($review);
                $this->updateDescription($review, $data, $this->services->get(ConnectionFactory::USER));
                $this->queueTask(
                    $review,
                    [ReviewModel::FIELD_DESCRIPTION => $review->getDescription(), self::IS_DESCRIPTION_CHANGE => true],
                    $oldValues
                );
            }
        } else {
            $translator = $this->services->get(TranslatorFactory::SERVICE);
            throw new InvalidArgumentException($translator->t('No description provided'));
        }
        return [ReviewModel::FIELD_DESCRIPTION => $review->getDescription()];
    }

   /**
     * It will take the review id check if review exist and perform validation on the participant
     * input data and update the participant data in review according to request.
     * @param mixed         $reviewId   the review id
     * @param array         $data       values from the request
     * @param array         $request    values from the request
     * @return array
     * @throws ForbiddenException
     * @throws \Exception
     */
    public function updateParticipants($reviewId, $data, $request)
    {
        $review = $this->fetch($reviewId, $this->services->get(ConnectionFactory::P4_ADMIN));
        $old    = $review->get();
        $this->checkAccess($review);
        $canEditReviewers         = $this->canEditReviewers($review);
        $messages                 = [];
        $defaultRetainedReviewers = [];
        $method                   = $request->getMethod();
        $projects                 = $review->getProjects();
        $p4Admin                  = $this->services->get(ConnectionFactory::P4_ADMIN);
        $defaultRetainedReviewers = UpdateService::mergeDefaultReviewersForProjects(
            $projects,
            $defaultRetainedReviewers,
            $p4Admin,
            [UpdateService::ALWAYS_ADD_DEFAULT => false]
        );
        if (!$canEditReviewers) {
            $translator = $this->services->get(TranslatorFactory::SERVICE);
            throw new ForbiddenException($translator->t('You do not have permission to edit reviewers.'));
        }
        $options                                      = [
            ParticipantsFilter::REVIEW => $review
        ];
        $filter                                       = $this->services->build(IParticipants::PARTICIPANTS, $options);
        $combinedReviewers                            = $filter->getCombinedReviewers($review, $data, $method);
        $data[ParticipantsFilter::COMBINED_REVIEWERS] = $combinedReviewers;
        $filter->setData($data);
        // indicate that we want to validate only fields present in $data
        $filter->setValidationGroupSafe(array_keys($data));
        if ($filter->isValid()) {
            $filtered = $filter->getValues();
            if ($method === HttpRequest::METHOD_DELETE) {
                $reviewerArrayToDelete = [];
                foreach ($filtered['combinedReviewers'] as $reviewer => $option) {
                    if (!array_key_exists($reviewer, $defaultRetainedReviewers)) {
                        $reviewerArrayToDelete[$reviewer][] = $option;
                    }
                }
                $reviewerArrayToSave =  array_diff_key($review->getParticipantsData(), $reviewerArrayToDelete);
                $review->setParticipantsData($reviewerArrayToSave);
            } else {
                $review->setParticipantsData($combinedReviewers);
            }
            $this->save($review);
        } else {
            $isRetainedReviewersError =  $filter->hasRetainedReviewersError();
            $retainReviewers          = $filter->hasRetainedReviewers();
            if ($isRetainedReviewersError) {
                $messages = $filter->getMessages();
                $values   = $filter->getValues();
                if (count($messages['combinedReviewers']) != count($retainReviewers)) {
                    throw new FilterException($filter);
                } else {
                    if ($method === HttpRequest::METHOD_DELETE) {
                        $reviewerArrayToDelete = [];
                        // removed all the default retain reviewer from delete
                        foreach ($values['combinedReviewers'] as $reviewer => $option) {
                            if (!array_key_exists($reviewer, $defaultRetainedReviewers)) {
                                $reviewerArrayToDelete[$reviewer][] = $option;
                            }
                        }
                        $reviewerArrayToSave =  array_diff_key($review->getParticipantsData(), $reviewerArrayToDelete);
                        $review->setParticipantsData($reviewerArrayToSave);
                        $this->save($review);
                    } elseif ($method === HttpRequest::METHOD_PUT || $method === HttpRequest::METHOD_POST) {
                        $reviewerArrayToPut = [];
                        // add all the default retain reviewer which are present in array
                        // so that only those retained reviewer will update which are upgrading
                        foreach ($defaultRetainedReviewers as $reviewer => $option) {
                            if (in_array($reviewer, $retainReviewers)) {
                                $values['combinedReviewers'][$reviewer] = $option;
                            }
                        }
                        foreach ($values['combinedReviewers'] as $reviewer => $option) {
                            if (isset($option[IParticipants::REQUIRED]) && empty($option[IParticipants::REQUIRED])) {
                                unset($option[IParticipants::REQUIRED]);
                            }
                            $reviewerArrayToPut[$reviewer] = $option;
                        }
                        foreach ($review->getParticipantsData() as $participant => $participantData) {
                            // Before we update participants data using put we need to preserve any fields
                            // already set (for example 'vote' and 'notificationsDisabled' etc)
                            foreach ($participantData as $participantField => $fieldValue) {
                                if ($participantField !== IParticipants::REQUIRED
                                    && $participantField !== ReviewModel::FIELD_MINIMUM_REQUIRED) {
                                    $reviewerArrayToPut[$participant][$participantField] = $fieldValue;
                                }
                            }
                        }
                        $review->setParticipantsData($reviewerArrayToPut);
                        $this->save($review);
                    }
                }
            } else {
                throw new FilterException($filter);
            }
        }
        $isReviewersChange = $old['participantsData'] != $review->getParticipantsData();
        if ($isReviewersChange) {
            $this->queueTask($review, [self::IS_REVIEWERS_CHANGE => true], $old);
        }
        return [$review, $messages];
    }

    /**
     * Format the input data to the required format so that it can be validated
     * from the ReviewFilter
     * @param $data
     * @return array
     * @throws \InvalidArgumentException
     * Sample Input
     * Array
     * (
     *      [participants] => Array
     *      (
     *          [users] => Array
     *          (
     *             [testUser] => Array([required] => yes)
     *          )
     *          [groups] => Array
     *          (
     *             [mygroup] => Array([required] => all)
     *          )
     *       )
     * )
     * Sample Output
     * Array
     * (
     *      [reviewers] => Array
     *          (
     *              [0] => testUser
     *              [1] => swarm-group-mygroup
     *          )
     *      [requiredReviewers] => Array
     *          (
     *              [0] => testUser
     *              [1] => swarm-group-mygroup
     *          )
     *      [reviewerQuorum] => Array
     *          ()
     * )
     */
    public function convertParticipantData($data)
    {
        $formatArray = [
            Review::REVIEWERS => [],
            Review::REQUIRED_REVIEWERS => [],
            Review::REVIEWER_QUORUMS => [],
        ];
        if (isset($data[IParticipants::PARTICIPANTS][IParticipants::USERS])) {
            $users = $data[IParticipants::PARTICIPANTS][IParticipants::USERS];
            foreach ($users as $key => $value) {
                $required = isset($value[IParticipants::REQUIRED]) ?
                    $value[IParticipants::REQUIRED] : IParticipants::NO;
                if ($required === IParticipants::YES) {
                    array_push($formatArray[Review::REVIEWERS], $key);
                    array_push($formatArray[Review::REQUIRED_REVIEWERS], $key);
                } elseif ($required === IParticipants::NO) {
                    array_push($formatArray[Review::REVIEWERS], $key);
                } else {
                    $translator = $this->services->get(TranslatorFactory::SERVICE);
                    throw new InvalidArgumentException(
                        $translator->t(
                            "'%s' must have a required value '%s' or '%s'",
                            [$key, IParticipants::YES, IParticipants::NO]
                        )
                    );
                }
            }
        }
        if (isset($data[IParticipants::PARTICIPANTS][IParticipants::GROUPS])) {
            $groups = $data[IParticipants::PARTICIPANTS][IParticipants::GROUPS];
            foreach ($groups as $key => $value) {
                    $required  = isset($value[IParticipants::REQUIRED]) ?
                        $value[IParticipants::REQUIRED] : IParticipants::NONE;
                    $groupName = GroupConfig::KEY_PREFIX .$key;
                if ($required === IParticipants::ONE) {
                    array_push($formatArray[Review::REVIEWERS], $groupName);
                    array_push($formatArray[Review::REQUIRED_REVIEWERS], $groupName);
                    $formatArray[Review::REVIEWER_QUORUMS][$groupName] = "1";
                } elseif ($required === IParticipants::ALL) {
                    array_push($formatArray[Review::REVIEWERS], $groupName);
                    array_push($formatArray[Review::REQUIRED_REVIEWERS], $groupName);
                } elseif ($required === IParticipants::NONE) {
                    array_push($formatArray[Review::REVIEWERS], $groupName);
                } else {
                    $translator = $this->services->get(TranslatorFactory::SERVICE);
                    throw new InvalidArgumentException(
                        $translator->t(
                            "'%s' must have a required value '%s' or '%s' or '%s'",
                            [$key, IParticipants::ALL, IParticipants::ONE, IParticipants::NONE,]
                        )
                    );
                }
            }
        }
        return $formatArray;
    }

    /**
     * Queue a review task with a type of 'review', an id from from the review and the data from the data parameter.
     * @param mixed         $review     the review
     * @param array|null    $data       data for the task. Defaults for the data are provided, callers can provide their
     * own to supplement/override. Defaults if not provided:
     * [
     *      'user'                => $this->services->get(ConnectionFactory::USER)->getId(),
     *      'isStateChange'       => $this->hasChanged(ReviewModel::FIELD_STATE, $values, $oldValues),
     *      'isAuthorChange'      => $this->hasChanged(ReviewModel::FIELD_AUTHOR, $values, $oldValues),
     *      'isVote'              => false,
     *      'isReviewersChange'   => false,
     *      'isDescriptionChange' => $this->hasChanged(ReviewModel::FIELD_DESCRIPTION, $values, $oldValues),
     *      'testStatus'          => null
     *      'deployStatus'        => null,
     *      'description'         => null,
     *      'previous'            => $oldValues,
     *      'quiet'               => false
     * ]
     * @param array|null    $oldValues  old values from the review
     * @return mixed|void
     */
    // Create to provide similar behaviour to the task queued in Reviews->IndexController->editReview for use when
    // v10 APIs result in data changing
    public function queueTask($review, $data = null, $oldValues = null)
    {
        $dataKeys  = array_keys($data);
        $oldValues = $oldValues ? $oldValues : [];
        $values    = $review->toArray();
        $data     += [
            ConnectionFactory::USER          =>
                in_array(ConnectionFactory::USER, $dataKeys)
                    ? $data[ConnectionFactory::USER]
                    : $this->services->get(ConnectionFactory::USER)->getId(),
            self::IS_STATE_CHANGE            =>
                in_array(self::IS_STATE_CHANGE, $dataKeys)
                    ? $data[self::IS_STATE_CHANGE]
                    : $this->hasChanged(ReviewModel::FIELD_STATE, $values, $oldValues),
            self::IS_AUTHOR_CHANGE           =>
                in_array(self::IS_AUTHOR_CHANGE, $dataKeys)
                    ? $data[self::IS_AUTHOR_CHANGE]
                    : $this->hasChanged(ReviewModel::FIELD_AUTHOR, $values, $oldValues),
            self::IS_VOTE                    => false,
            self::IS_REVIEWERS_CHANGE        => false,
            self::IS_DESCRIPTION_CHANGE      =>
                in_array(self::IS_DESCRIPTION_CHANGE, $dataKeys)
                    ? $data[self::IS_DESCRIPTION_CHANGE]
                    : $this->hasChanged(ReviewModel::FIELD_DESCRIPTION, $values, $oldValues),
            ReviewModel::FIELD_TEST_STATUS   => null,
            ReviewModel::FIELD_DEPLOY_STATUS => null,
            ReviewModel::FIELD_DESCRIPTION   => null,
            self::PREVIOUS                   => $oldValues,
            self::QUIET                      => false
        ];

        $queue = $this->services->get(QueueManager::SERVICE);
        $queue->addTask(ListenerFactory::REVIEW, $review->getId(), $data);
    }


    /**
     * Can the files be archived and downloaded.
     *
     * @param int $change the changelist id we want to fetch
     * @return bool
     */
    public function canArchive($change)
    {
        $archiver = $this->services->get(Services::ARCHIVER);
        if ($archiver->canArchive()) {
            $p4Admin    = $this->services->get(ConnectionFactory::P4_ADMIN);
            $changeDAO  = $this->services->get(IModelDAO::CHANGE_DAO);
            $changelist = $changeDAO->fetchById($change, $p4Admin);
            $files      = $changelist->getFileData($changelist->isPending());
            if (count($files)) {
                foreach ($files as $file) {
                    if ($file[ DecodeSpec::TYPE ] !== DecodeSpec::STREAM) {
                        return true;
                    }
                }
            }
        }
        return false;
    }
    /**
     * This method implements our access controls for who can/can-not edit the reviewers.
     * Note joining/leaving and adjusting your own required'ness is a separate operation
     * and not governed by this check.
     *
     * @param   ReviewModel  $review     review model
     * @return  bool    true if active user can edit the review, false otherwise
     * @throws \Exception
     */
    private function canEditReviewers(ReviewModel $review)
    {
        $permissions = $this->services->get(Permissions::PERMISSIONS);

        // if you aren't authenticated you aren't allowed to edit
        if (!$permissions->is(Permissions::AUTHENTICATED)) {
            return false;
        }

        // if you are an admin, the author, or there aren't any
        // projects associated with this review, you can edit.
        $userId = $this->services->get(ConnectionFactory::USER)->getId();
        if ($permissions->is('admin')
            || $review->get('author') === $userId
            || !$review->getProjects()
        ) {
            return true;
        }

        // looks like this review impacts projects, let's figure
        // out who the involved members and moderators are.
        $p4Admin    = $this->services->get(ConnectionFactory::P4_ADMIN);
        $members    = [];
        $moderators = [];
        $impacted   = $review->getProjects();
        $projectDAO = $this->services->get(IModelDAO::PROJECT_DAO);
        $projects   = $projectDAO->fetchAll(['ids' => array_keys($impacted)], $p4Admin);
        foreach ($projects as $project) {
            $branches   = $impacted[$project->getId()];
            $moderators = array_merge($moderators, $project->getModerators($branches));
            $members    = array_merge($members,    $project->getAllMembers());
        }

        // if there are moderators, only they can edit.
        // if there aren't any moderators, then any members can edit.
        return $p4Admin->stringMatches($userId, $moderators ?: $members);
    }

    /**
     * Work out if the content of a review has changed. If this is the first revision it is always changed otherwise
     * a change service is used to determine any differences between the head change and the change on the previous
     * revision.
     * @param ReviewModel $review
     * @return bool
     * @throws Exception
     */
    public function hasContentChanged(ReviewModel $review)
    {
        $head = $review->getHeadVersion();
        if ($head === 1) {
            $changed = true;
        } else {
            $previousChange = $review->getChangeOfVersion($head - 1, true);
            $currentChange  = $review->getHeadChange(true);
            $changeService  = $this->services->get(Services::CHANGE_SERVICE);
            $p4Admin        = $this->services->get(ConnectionFactory::P4_ADMIN);
            $changed        = $changeService->hasContentChanged($p4Admin, $currentChange, $previousChange);
        }
        return $changed;
    }

    /**
     * Return review record without any permission and private project filter check.
     * @param int                       $id         id of the record
     * @param ConnectionInterface|null  $connection the connection
     * @return mixed
     * @throws RecordNotFoundException
     */
    public function fetchNoCheck($id, ConnectionInterface $connection = null)
    {
        return ReviewModel::fetch($id, $this->getConnection($connection));
    }
}
