<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2020.1/2030377
 */

namespace TestIntegration\Service;

use Api\IRequest;
use Application\Config\ConfigException;
use Application\Config\ConfigManager;
use Application\Config\IDao;
use Application\Connection\ConnectionFactory;
use Application\I18n\TranslatorFactory;
use Application\Log\SwarmLogger;
use Application\View\Helper\ViewHelperFactory;
use Application\Escaper\Escaper;
use Files\MimeType;
use Interop\Container\ContainerInterface;
use P4\Model\Fielded\Iterator;
use Projects\Model\Project as ProjectModel;
use Record\Key\AbstractKey;
use TestIntegration\Controller\TestRunApi;
use TestIntegration\Filter\EncodingValidator;
use TestIntegration\Filter\ITestRun;
use TestIntegration\Filter\StatusValidator;
use TestIntegration\Model\ITestDefinition;
use TestIntegration\Model\ITestRun as ITestRunModel;
use TestIntegration\Model\TestRun;
use Laminas\Http\Client as HttpClient;
use Laminas\Http\Request;
use Laminas\Http\Response;
use Workflow\Model\IWorkflow;

/**
 * Class TestExecutor
 * @package TestIntegration\Service
 */
class TestExecutor implements ITestExecutor
{
    private $services;
    private $logger;
    private $testRunDao;
    protected $review;
    private $affectedProjects;
    private $projects;
    private $testsForWorkflows;
    private $hasContentChanged;
    private $p4Admin;

    // The minimum API version that will support the new callback urls
    const API_VERSION = 'v10';

    use ProjectTestNameTrait;

    /**
     * TestExecutor constructor.
     * @inheritDoc
     */
    public function __construct(ContainerInterface $services, array $options = null)
    {
        $this->services   = $services;
        $this->logger     = $services->get(SwarmLogger::SERVICE);
        $this->testRunDao = $services->get(IDao::TEST_RUN_DAO);
        $this->p4Admin    = $services->get(ConnectionFactory::P4_ADMIN);
    }

    /**
     * @inheritDoc
     */
    public function runReviewTests(AbstractKey $review, array $affectedProjects, Iterator $projects)
    {
        $this->review            = $review;
        $this->affectedProjects  = $affectedProjects;
        $this->projects          = $projects;
        $this->hasContentChanged = $this->services->get(IDao::REVIEW_DAO)->hasContentChanged($review);
        $testsForWorkflows       = $this->getTestsForWorkflows();
        if ($this->hasContentChanged) {
            // Content has changed - we run all tests (subject to event/pending) regardless of the previous test runs
            $this->runWorkflowTests($testsForWorkflows);
            $this->runProjectTests();
        } else {
            // No content change - we may still need to run tests but we can exclude those that passed on the previous
            // revision of the review
            $previouslyPassed = [];
            $head             = $review->getHeadVersion();
            if ($head > 1) {
                $testRunIds = $review->getTestRuns($head - 1);
                if ($testRunIds) {
                    foreach ($testRunIds as $testRunId) {
                        $testRun = $this->testRunDao->fetch($testRunId, $this->p4Admin);
                        if ($testRun->getStatus() === StatusValidator::STATUS_PASS) {
                            $previouslyPassed[] = $testRun;
                        }
                    }
                }
            }
            $previousPassedNames = array_map(
                function ($testRun) {
                    return $testRun->getTest();
                },
                $previouslyPassed
            );
            $this->runWorkflowTests($testsForWorkflows, $previousPassedNames);
            $this->runProjectTests($previousPassedNames);
            $this->cloneTestRuns($review, $previouslyPassed, $previousPassedNames);
        }
    }

    /**
     * Copies previously passed test run results to a new test run and adds it to the latest revision of the review.
     * All the runs in 'previouslyPassed' are candidates for copy but they will only be copied if the test name is
     * also in 'previousPassedNames'. The names may have been manipulated due to onSubmit tests not causing a clone.
     * @param mixed     $review                 the review
     * @param array     $previouslyPassed       previously passed test runs
     * @param array     $previousPassedNames    previously passed test runs names.
     */
    private function cloneTestRuns($review, $previouslyPassed, $previousPassedNames)
    {
        $translator = $this->services->get(TranslatorFactory::SERVICE);
        foreach ($previouslyPassed as $testRun) {
            if (in_array($testRun->getTest(), $previousPassedNames)) {
                $newRun = new TestRun;
                $newRun->set($testRun->toArray());
                $newRun = $newRun->setId(null)->setTitle($testRun->getTitle() . ' (' . $translator->t('Copy') . ')');
                $newRun = $this->testRunDao->save($newRun);
                $review = $review->addTestRun($newRun->getId());
            }
        }
        $reviewDao = $this->services->get(IDao::REVIEW_DAO);
        $reviewDao->save($review);
    }

    /**
     * Gets all the tests to run based on the global workflow and any workflows associated with projects and branches
     * @return mixed
     */
    public function getTestsForWorkflows()
    {
        $projectDao  = $this->services->get(IDao::PROJECT_DAO);
        $workflows   = $projectDao->getWorkflowsForAffectedProjects($this->affectedProjects);
        $workflows[] = IWorkflow::GLOBAL_WORKFLOW_ID;
        $workflowDao = $this->services->get(IDao::WORKFLOW_DAO);
        return $workflowDao->getTestsForWorkflows($workflows, [], $this->p4Admin);
    }

    /**
     * @inheritDoc
     */
    public function runWorkflowTests($testsForWorkflows, &$previouslyPassedNames = [])
    {
        $review  = $this->review;
        $testIds = $testsForWorkflows
            ? isset($testsForWorkflows[IWorkflow::EVENT_ON_UPDATE])
                ? $testsForWorkflows[IWorkflow::EVENT_ON_UPDATE] : []
            : [];
        // Remove previously passed 'onUpdate' tests.
        // 'onSubmit' will be checked next and they always run
        $testIds = array_diff($testIds, $previouslyPassedNames);
        if (!$review->isPending()) {
            $onSubmitTestIds = $testsForWorkflows
                ? isset($testsForWorkflows[IWorkflow::EVENT_ON_SUBMIT])
                    ? $testsForWorkflows[IWorkflow::EVENT_ON_SUBMIT] : []
                : [];
            // Remove onSubmit ids from previously passed so that we do not clone them
            $previouslyPassedNames = array_diff($previouslyPassedNames, $onSubmitTestIds);
            $testIds               = array_unique(array_merge($testIds, $onSubmitTestIds));
        }
        // Get the enabled test definitions
        if ($testIds) {
            $testDefinitionDAO = $this->services->get(IDao::TEST_DEFINITION_DAO);
            $testDefinitions   = $testDefinitionDAO->fetchAll(
                [AbstractKey::FETCH_BY_IDS => $testIds],
                $this->p4Admin
            );
            // Set review vars that will not change within the loop
            $change       = $review->getHeadChange(true);
            $status       = $review->isPending() ? 'shelved' : 'submitted';
            $reviewId     = $review->getId();
            $version      = $review->getHeadVersion();
            $reviewToken  = $review->getToken();
            $reviewStatus = $review->getState();
            $description  = $review->getDescription();
            $projectInfo  = $this->getProjectInfo();


            // Run a test for each test definition
            foreach ($testDefinitions as $testDefinition) {
                // Create the test run record
                $testRun = $this->createTestRun($testDefinition->getId(), $testDefinition->getName());

                // Test run could not be created, so we continue on
                if (is_null($testRun)) {
                    continue;
                }

                // Get the callback urls
                $testRunId    = $testRun->getId();
                $callbackUrls = $this->generateCallbackUrls($testRunId, $reviewToken);

                // Used to build out the CI request url
                $fields = [
                    '{change}'       => $change,
                    '{status}'       => $status,
                    '{review}'       => $reviewId,
                    '{version}'      => $version,
                    '{test}'         => $testDefinition->getName(),
                    '{testRunId}'    => $testRunId,
                    '{projects}'     => $projectInfo['projectIds'],
                    '{projectNames}' => $projectInfo['projectNames'],
                    '{branches}'     => $projectInfo['projectAndBranchIds'],
                    '{branch}'       => $projectInfo['branchIds'],
                    '{branchName}'   => $projectInfo['branchNames'],
                    '{pass}'         => $callbackUrls['pass'],
                    '{fail}'         => $callbackUrls['fail'],
                    '{update}'       => $callbackUrls['update'],
                    '{reviewStatus}' => $reviewStatus,
                    '{description}'  => $description
                ];

                $this->runTest($testDefinition, $fields, $testRun);
            }
        }
    }

    /**
     * @inheritDoc
     */
    public function runProjectTests($previouslyPassedNames = [])
    {
        $review = $this->review;

        // Set review vars that will not change within the loop
        $change         = $review->getHeadChange(true);
        $status         = $review->isPending() ? 'shelved' : 'submitted';
        $reviewId       = $review->getId();
        $version        = $review->getHeadVersion();
        $reviewToken    = $review->getToken();
        $description    = $review->getDescription();
        $reviewStatus   = $review->getState();
        $testStartTimes = [];

        // reviews can impact multiple projects and each project can have its own test config
        // note we only include projects/branches the change being processed impacts.
        foreach ($this->affectedProjects as $projectId => $branchIds) {
            // get the full project object and the list of impacted branch names
            $project         = isset($this->projects[$projectId]) ? $this->projects[$projectId] : null;
            $branchNames     = $this->getBranchNamesFromIds($project, $branchIds);
            $projectTestName = $this->getProjectTestName($projectId);
            // if enabled, configured and not previously passed, run automated tests
            if ($project
                && $project->getTests('enabled')
                && $project->getTests('url')
                && !in_array($projectTestName, $previouslyPassedNames)) {
                $testRun = $this->createTestRun($projectTestName, $project->getName());

                // Test run could not be created, so we continue on
                if (is_null($testRun)) {
                    continue;
                }

                // Get the callback urls
                $testRunId    = $testRun->getId();
                $callbackUrls = $this->generateCallbackUrls($testRunId, $reviewToken);

                $fields = [
                    '{change}'        => $change,
                    '{status}'        => $status,
                    '{review}'        => $reviewId,
                    '{version}'       => $version,
                    '{test}'          => $testRun->getTest(),
                    '{testRunId}'     => $testRunId,
                    '{project}'       => $project->getId(),
                    '{projectName}'   => $project->getName(),
                    '{branch}'        => implode(',', $branchIds),
                    '{branchName}'    => implode(',', $branchNames),
                    '{pass}'          => $callbackUrls['pass'],
                    '{fail}'          => $callbackUrls['fail'],
                    '{update}'        => $callbackUrls['update'],
                    '{reviewStatus}'  => $reviewStatus,
                    '{description}'   => $description
                ];

                 $this->runProjectTest($project, $fields, $testRun, $testStartTimes);
            }

            // if enabled and configured, run deploy
            if ($project && $project->getDeploy('enabled') && $project->getDeploy('url')) {
                // compose deploy success/fail callback urls.
                $urlHelper        = $this->services->get('ViewHelperManager')->get(ViewHelperFactory::QUALIFIED_URL);
                $deploySuccessUrl = $urlHelper(
                    'review-deploy',
                    ['review' => $review->getId(), 'token' => $review->getToken(), 'status' => 'success']
                );
                $deployFailUrl    = $urlHelper(
                    'review-deploy',
                    ['review' => $review->getId(), 'token' => $review->getToken(), 'status' => 'fail']
                );
                // customize test url for this project.
                $search  = [
                    '{change}',
                    '{status}',
                    '{review}',
                    '{project}',
                    '{projectName}',
                    '{branch}',
                    '{branchName}',
                    '{success}',
                    '{fail}'
                ];
                $replace = array_map(
                    'rawurlencode',
                    [
                        $change,
                        $status,
                        $reviewId,
                        $project->getId(),
                        $project->getName(),
                        implode(',', $branchIds),
                        implode(',', $branchNames),
                        $deploySuccessUrl,
                        $deployFailUrl
                    ]
                );

                $url      = str_replace($search, $replace, $project->getDeploy('url'));
                $response = $this->doRequest($url, 'automated deploy', EncodingValidator::URL);
                // All good, nothing to do here, so exit early
                if ($response && $response->isSuccess()) {
                    $review->set('deployStatus', 'deploying');
                } elseif (is_null($response)) {
                    $redactUrl = $this->redactUrl($url);
                    $review->set(
                        'deployStatus',
                        'failed:'.sprintf("There was no response from %s", explode('?', $redactUrl)[0])
                    );
                } else {
                    $review->set(
                        'deployStatus',
                        'failed:'.sprintf("%s:%s", $response->getStatusCode(), $response->getReasonPhrase())
                    );
                }
                $review->save();
            }
        }

        $details = $review->getTestDetails(true);
        if ($testStartTimes || count($details['startTimes']) > count($details['endTimes'])) {
            $review->setTestDetails(['startTimes' => $testStartTimes, 'endTimes' => []] + $details)->save();
        }
    }

    /**
     * @inheritDoc
     */
    public function runTest(ITestDefinition $testDefinition, array $fields, ITestRunModel $testRun)
    {
        // We never allow the description in the url, so we preemptively remove it
        $url        = str_replace('{description}', '', $testDefinition->getUrl());
        $bodyFields = $fields;
        // Build out the necessary parameters for making the CI request
        $search   = array_keys($fields);
        $encoding = $testDefinition->getEncoding();
        // We treat fields in the body differently than fields in the URL itself. Expansions in the URL are always
        // fully escaped, encoding on body expansions depends on the request type (url encoded or json encoded)
        $this->encodeBodyExpansions($bodyFields, $encoding);
        $replace  = array_map('rawurlencode', array_values($fields));
        $url      = str_replace($search, $replace, $url);
        $postBody = str_replace($search, array_values($bodyFields), trim($testDefinition->getBody()));
        $headers  = $testDefinition->getHeaders() ?? [];
        $this->logger->info('Tests triggered for review: ' . $this->review->getId() . ' URL: ' . $url);
        $request  = [
            $url,
            $testDefinition->getDescription(),
            $encoding,
            $postBody,
            $headers,
            $testDefinition->getTimeout()
        ];
        $response = $this->doRequest(...$request);
        $this->validateResponse($testRun, $url, $response);
        return $request;
    }

    /**
     * @inheritDoc
     */
    public function runProjectTest(ProjectModel $project, array $fields, ITestRunModel $testRun, array &$testStartTimes)
    {
        $review = $this->review;
        // We never allow the description in the url, so we preemptively remove it
        $url        = str_replace('{description}', '', $project->getTests('url'));
        $bodyFields = $fields;
        // Build out the necessary parameters for making the CI request
        $search     = array_keys($fields);
        $postFormat = $project->getTests('postFormat') ?? EncodingValidator::URL;
        // We treat fields in the body differently that fields in the URL itself. Expansions in the URL are always
        // fully escaped, encoding on body expansions depends on the request type (url encoded or json encoded)
        $this->encodeBodyExpansions($bodyFields, $postFormat);
        $replace     = array_map('rawurlencode', array_values($fields));
        $url         = str_replace($search, $replace, $url);
        $postBody    = str_replace($search, array_values($bodyFields), trim($project->getTests('postBody')));
        $description = "Tests for " . $project->getId();
        $this->logger->info('Tests triggered for review: ' . $review->getId() . ' URL: ' . $url);
        $request  = [$url, $description, $postFormat, $postBody];
        $response = $this->doRequest($url, $description, $postFormat, $postBody);
        if ($response && $response->isSuccess()) {
            $testStartTimes[] = time();
            $review->set('testStatus', StatusValidator::STATUS_RUNNING);
            $review->save();
        }
        $this->validateResponse($testRun, $url, $response);
        return $request;
    }

    /**
     * Processes the currentProjects and projects arrays to return lists of project ids & names and branch ids (which
     * are composed from both the project id and the branch id). Also history branch ids and branch name without
     * prefix of project to enable old project tests to be transferred into new world.
     *
     * @return array
     * @throws ConfigException
     */
    protected function getProjectInfo()
    {
        $projectIds          = [];
        $projectNames        = [];
        $projectAndBranchIds = [];
        $branchIds           = [];
        $branchNames         = [];
        $config              = $this->services->get(ConfigManager::CONFIG);
        $separator           = ConfigManager::getValue(
            $config,
            ConfigManager::TEST_DEFINITIONS_PROJECT_AND_BRANCH_SEPARATOR,
            ':'
        );


        foreach ($this->affectedProjects as $projectId => $currentBranchIds) {
            if (isset($this->projects[$projectId])) {
                $project        = $this->projects[$projectId];
                $projectIds[]   = $projectId;
                $projectNames[] = $project->getName();

                foreach ($currentBranchIds as $branchId) {
                    // Construct the branch id in the form of <projectId>:<branchId>
                    $projectAndBranchIds[] = sprintf('%s'.$separator.'%s', $projectId, $branchId);
                    $branchIds[]           =  $branchId;
                    $branchNames[]         =  $project->getBranch($branchId)['name'];
                }
            }
        }
        return [
            'projectIds'          => implode(',', $projectIds),
            'projectNames'        => implode(',', $projectNames),
            'projectAndBranchIds' => implode(',', $projectAndBranchIds),
            'branchIds'           => implode(',', $branchIds),
            'branchNames'         => implode(',', $branchNames)
        ];
    }

    /**
     * Convenience method to get a list of branch names, given branch ids
     *
     * @param ProjectModel $project
     * @param array        $branchIds
     *
     * @return array
     */
    protected function getBranchNamesFromIds(ProjectModel $project, array $branchIds)
    {
        $branches    = $project->getBranches();
        $branchNames = [];
        foreach ($branchIds as $branchId) {
            foreach ($branches as $branch) {
                if ($branch['id'] == $branchId) {
                    $branchNames[] = $branch['name'];
                    break;
                }
            }
        }

        return $branchNames;
    }

    /**
     * Creates a new TestRun instance
     * @param string    $testIdentifier     identifier for the test - either the test definition id or the project test
     *                                      identifier based on the project id.
     * @param string    $title              the title for the test
     * @return mixed
     */
    protected function createTestRun(string $testIdentifier, string $title)
    {
        $review  = $this->review;
        $testRun = null;
        $data    = [
            TestRun::FIELD_CHANGE     => $review->getId(),
            TestRun::FIELD_VERSION    => $review->getHeadVersion(),
            TestRun::FIELD_TEST       => $testIdentifier,
            TestRun::FIELD_START_TIME => time(),
            TestRun::FIELD_STATUS     => StatusValidator::STATUS_RUNNING,
            TestRun::FIELD_UUID       => $review->getToken(),
            TestRun::FIELD_TITLE      => $title
        ];

        $filter = $this->services->get(ITestRun::NAME);
        $filter->setData($data);

        if ($filter->isValid()) {
            $testRun = new TestRun();
            $testRun = $testRun->set($data);
            $this->testRunDao->save($testRun);
            $review->addTestRun($testRun->getId())->save();
        } else {
            $errors = $filter->getMessages();
            $this->logValidationErrors($errors);
        }

        return $testRun;
    }

    /**
     * Executes the actual CI request
     *
     * @param string   $url          url for the CI system to run the test
     * @param string   $description  description of the test
     * @param string   $postBody     request parameters that will be sent in the body or in the url
     * @param string   $postFormat   either url or json
     * @param array    $headers      headers, if specified
     * @param int      $timeout      optional timeout, defaults to 0 and is ignored if 0
     *
     * @return mixed
     */
    protected function doRequest(
        string $url,
        string $description,
        string $postFormat,
        string $postBody = '',
        array $headers = [],
        int $timeout = 0
    ) {
        $response = null;
        $logger   = $this->logger;

        // extract the http client options; including any special overrides for our host
        $options = $this->services->get(ConfigManager::CONFIG) + ['http_client_options' => []];
        $options = (array)$options['http_client_options'];

        // below function call is to encode the url path
        $url = $this->encodeUrlPath($url);
        
        // attempt a request for the given url to trigger tests.
        try {
            $client = new HttpClient;
            $client->setUri($url);
            $client->setHeaders($headers);

            // if we have post data, ensure we make a POST request
            if (!empty($postBody)) {
                $client->setMethod(Request::METHOD_POST);

                // parse body based on its format (url, xml or json)
                switch ($postFormat) {
                    case strcasecmp($postFormat, EncodingValidator::URL) === 0:
                        parse_str($postBody, $postParams);
                        $client->setParameterPost($postParams);
                        break;
                    default:
                        $client->setEncType(MimeType::$mimeTypes[strtolower($postFormat)]);
                        $client->setRawBody($postBody);
                        break;
                }
            }

            // calculate options, including host based overrides, and set them
            if (isset($options['hosts'][$client->getUri()->getHost()])) {
                $options = (array)$options['hosts'][$client->getUri()->getHost()] + $options;
            }
            unset($options['hosts']);
            if ($timeout > 0) {
                $options['timeout'] = $timeout;
            }
            $client->setOptions($options);
            $logger->trace(get_class($this) . ': test dispatch request ' . var_export($client->getRequest(), true));
            // attempt trigger remote build.
            $response = $client->dispatch($client->getRequest());
        } catch (\Exception $e) {
            $logger->err($e);
        }
        return $response;
    }

    /**
     * Generate callback urls which can be used by the CI system to update a given test run
     *
     * @param int      $testRunId   id of the test run
     * @param string   $uuid        uuid of the test run
     *
     * @return array
     */
    protected function generateCallbackUrls($testRunId, $uuid)
    {
        $urlHelper = $this->services->get('ViewHelperManager')->get(ViewHelperFactory::QUALIFIED_URL);
        $params    = [
            IRequest::VERSION   => self::API_VERSION,
            TestRun::FIELD_ID   => $testRunId,
            TestRun::FIELD_UUID => $uuid
        ];

        return [
            'pass'   => $urlHelper(TestRunAPI::PASS_URL_ROUTE, $params),
            'fail'   => $urlHelper(TestRunAPI::FAIL_URL_ROUTE, $params),
            'update' => $urlHelper(TestRunAPI::UPDATE_NOAUTH_URL_ROUTE, $params)
        ];
    }

    /**
     * Handle the logging of filter validation errors
     * @param array $errors
     */
    protected function logValidationErrors(array $errors)
    {
        $message = "TestRun could not be created!\n";

        foreach ($errors as $field => $fieldErrors) {
            $message .= "$field => ";
            foreach ($fieldErrors as $type => $error) {
                $message .= "$type: $error\n";
            }
        }

        $this->logger->err($message);
    }

    /**
     * Ensure that the response from the CI system is valid
     * If the response is not valid we:
     *  - log that there was no response and if there was a response, add the reason and code
     *  - set the TestRun record's status as a fail
     *  - set the TestRun record's messages as 1-element array, containing a string indicating the code & reason for
     *    the validation failure or a message saying there was no response
     *
     * @param ITestRunModel   $testRun    TestRun record
     * @param string          $url        url of the CI system
     * @param Response|null   $response   response from the CI system or null
     */
    protected function validateResponse(ITestRunModel $testRun, string $url, $response)
    {
        // All good, nothing to do here, so exit early
        if ($response && $response->isSuccess()) {
            return;
        }
        if (is_null($response)) {
            $redactUrl = $this->redactUrl($url);
            $message   = sprintf("There was no response from %s", explode('?', $redactUrl)[0]);
        } else {
            $message = sprintf("%s:%s", $response->getStatusCode(), $response->getReasonPhrase());
        }
        $this->logger->err(sprintf("Failed to trigger %s: %s, %s", $testRun->getTest(), $url, $message));
        $testRun->setStatus(StatusValidator::STATUS_FAIL);
        $testRun->setMessages([$message]);
        $this->testRunDao->save($testRun);
    }

    /**
     * Escapes a string to deal with markdown, newlines and unicode
     * @param string   $dirty   string to be cleaned
     * @return string
     */
    protected function sanitize(string $dirty)
    {

        if (empty($dirty)) {
            $clean = '';
        } else {
            $parsedown = new \Parsedown();
            $markdown  = $parsedown->text($dirty);
            $stripped  = strip_tags($markdown);
            $clean     = htmlentities($stripped);
        }
        return $clean;
    }

    /**
     * As we don't want to show the user login or token we remove them from the url
     * We also remove query and hashes from the url as well.
     * @param string $url The url we send to test system.
     * @return string without user credentials, query fields and hashes.
     */
    protected function redactUrl(string $url)
    {
        $parts = parse_url($url);
        return $parts['scheme'] . '://' .
            $parts['host'] .
            (isset($parts['port']) ? ':'.$parts['port'] : '') .
            (isset($parts['path']) ? $parts['path'] : '');
    }

    /**
     * It encodes the url path and return the new url
     * @param string   $url url for the CI system to run the test
     * @return string
     */
    protected function encodeUrlPath(string $url)
    {
        $escaper = new Escaper();
        $url     = $escaper->escapeFullUrl(rawurldecode($url));
        return $url;
    }

    /**
     * Encode field values for expansions
     * @param array         $fields     key of expansion field to value
     * @param string        $encoding   whether we are dealing with URL encoding or JSON
     * @return bool
     */
    protected function encodeBodyExpansions(&$fields, $encoding)
    {
        return array_walk($fields, [$this, 'encodeBodyExpansion'], $encoding);
    }

    /**
     * Encode a value according to the encoding type.
     * In a JSON encoded request we just want to encode any field expansion to be JSON safe rather than fully escaping
     * In a URL encoded request we fully escape fields in the body. For description we also strip tags and html entities
     * @param $value
     * @param $key
     * @param $encoding
     */
    protected function encodeBodyExpansion(&$value, $key, $encoding)
    {
        switch ($encoding) {
            case strcasecmp(EncodingValidator::JSON, $encoding) === 0:
                $value = trim(json_encode($value, JSON_UNESCAPED_SLASHES), '"');
                break;
            default:
                if ($key === '{description}') {
                    $value = $this->sanitize($value);
                }
                $value = rawurlencode($value);
                break;
        }
    }
}
